from inspect import getmodule as _getmodule
from inspect import stack as _stack
from logging import DEBUG as _DEBUG
from logging import INFO as _INFO
from logging import WARN as _WARN
from logging import Logger as _Logger
from logging import StreamHandler as _StreamHandler
from logging import getLogger as _getLogger
from typing import Dict as _Dict
from typing import Iterable as _Iterable
from typing import List as _List
from typing import Optional as _Optional

from pythonjsonlogger import jsonlogger as _jsonlogger


def get_logger() -> _Logger:
    """Acquires a logger."""
    stack = _stack()
    caller = stack[1]

    module = _getmodule(caller[0])
    assert module
    module_name = module.__name__
    function_name = caller[3]
    logger_name = f"{module_name}.{function_name}"
    return LOGGERS.get(logger_name)


def set_logging_to_debug() -> None:
    LOGGERS.update_options(LoggerOptions().set_level(_DEBUG))


def set_logging_to_info() -> None:
    LOGGERS.update_options(LoggerOptions().set_level(_INFO))


def set_logging_to_warn() -> None:
    LOGGERS.update_options(LoggerOptions().set_level(_WARN))


class LoggerOptions:
    def __init__(self: "LoggerOptions") -> None:
        self.level = _INFO

    def set_level(self: "LoggerOptions", level: int) -> "LoggerOptions":
        self.level = level
        return self

    def update_logger(self: "LoggerOptions", logger: _Logger) -> None:
        logger.setLevel(self.level)
        for handler in logger.handlers:
            handler.setLevel(self.level)


DEFAULT_LOGGER_OPTIONS = LoggerOptions()


class Loggers:
    def __init__(self: "Loggers", options: _Optional[LoggerOptions] = None) -> None:
        self._loggers: _Dict[str, _Logger] = {}
        self._options = options if options else DEFAULT_LOGGER_OPTIONS

    def get(self: "Loggers", name: str) -> _Logger:
        if name not in self._loggers:
            self._loggers[name] = self._new(name)
        return self._loggers[name]

    def _new(self: "Loggers", name: str) -> _Logger:
        return config_logger(_getLogger(name), self._options)

    def update_options(self: "Loggers", options: LoggerOptions) -> None:
        self._options = options
        for logger in self._loggers.values():
            options.update_logger(logger)


def config_logger(logger: _Logger, options: LoggerOptions) -> _Logger:
    logger.setLevel(options.level)

    console_handler = _StreamHandler()
    console_handler.setLevel(options.level)

    custom_format = " ".join(_log_format(LEAN_KEYS))
    formatter = _jsonlogger.JsonFormatter(custom_format)
    console_handler.setFormatter(formatter)

    logger.addHandler(console_handler)
    return logger


def _log_format(keys: _Iterable[str]) -> _List[str]:
    return [f"%({key:s})s" for key in keys]


LOGGERS = Loggers()
LEAN_KEYS = [
    "asctime",
    "levelname",
    "message",
    "funcName",
    "pathname",
]

DEFAULT_KEYS = [
    "asctime",
    "filename",
    "funcName",
    "levelname",
    "levelno",
    "lineno",
    "module",
    "message",
    "name",
    "pathname",
    "relativeCreated",
    "thread",
    "threadName",
]



"""
This module contains the environment variables used by the application.
"""

from os import environ as _environ
from os.path import expanduser as _expanduser
from os.path import join as _join

from dotenv import load_dotenv as _load_dotenv

_load_dotenv()


def root_folder_path() -> str:
    """
    Get the root folder path for the data storage.
    """
    if "ROOT_FOLDER_PATH" in _environ:
        return _environ["ROOT_FOLDER_PATH"]
    return _join(_expanduser("~"), "data")


def writer_folder_path() -> str:
    """
    Get the folder path for the data writer.
    """
    if "FILE_WRITER_PATH" in _environ:
        return _environ["FILE_WRITER_PATH"]
    return root_folder_path()


def reader_folder_path() -> str:
    """
    Get the folder path for the data reader.
    """
    if "FILE_READER_PATH" in _environ:
        return _environ["FILE_READER_PATH"]
    return root_folder_path()


def vault_token() -> str:
    """
    Get the vault token.
    """
    if "VAULT_TOKEN" in _environ:
        return _environ["VAULT_TOKEN"]
    return ""



from typing import Any as _Any
from typing import List as _List
from typing import Tuple as _Tuple
from typing import Union as _Union

from ..fs.writer import FileWriter as _FileWriter
from ..importer.importer import Importer as _Importer
from ..logger import get_logger as _get_logger

LOGGER = _get_logger()


class Fetcher:
    """
    Fetches the given list of urls using the given importer and stores the result using the given
    writer.

    TODO: parse the domain from the url to set the prefix
    """

    def __init__(
        self: "Fetcher",
        importer: _Importer,
        writer: _FileWriter,
        urls: _List[_Union[str, _Tuple[str, _Any, _Any]]],
        prefix: str = "",
    ) -> None:
        self.importer = importer
        self.writer = writer
        self.urls = urls
        self.prefix = prefix

    def fetch(self: "Fetcher") -> None:
        for url in self.urls:
            if isinstance(url, str):
                args = []
                kwargs = {}
            else:
                args = url[1]
                kwargs = url[2]
                url = url[0]
            LOGGER.debug(f"Fetching '{url}'")
            self.writer.write(self.get_name(url), self.importer.get(url, *args, **kwargs))

    def get_name(self: "Fetcher", url: str) -> str:
        return self.prefix + url



"""Find and replace utilities."""

from collections import namedtuple
from typing import Callable, Iterable, Union

from pandas import DataFrame, Index, Series

ExactMatch = namedtuple("ExactMatch", "column value")
Replacement = namedtuple("Replacement", "column value")
FindCallBack = Callable[[DataFrame], Series]
MatcherT = Iterable[Union[ExactMatch, FindCallBack]]


def find(df: DataFrame, matchers: MatcherT) -> DataFrame:
    """Find all rows where the columns exacly match the values."""
    for matcher in matchers:
        if isinstance(matcher, ExactMatch):
            column, value = matcher
            df = df[df[column] == value]
        elif callable(matcher):
            df = df[matcher(df)]
        else:
            raise RuntimeError(f"Unable to process matcher '{matcher}'")
    return df


def finder(exact_match: MatcherT) -> Callable[[DataFrame], DataFrame]:
    """Make a funciton that will find all rows from a dataframe where the columns exactly match the given values."""

    def _find(df: DataFrame) -> DataFrame:
        return find(df, exact_match)

    return _find


def replace(df: DataFrame, index: Index, replacements: Iterable[Replacement]) -> DataFrame:
    """Replace the specified columns with specified values at the given index."""
    for column, value in replacements:
        df.loc[index, [column]] = value
    return df


def replacer(
    replacements: Iterable[Replacement],
) -> Callable[[DataFrame, Index], DataFrame]:
    """Make a funtion to replace the specified columns with specified values at the given index."""

    def _replace(df: DataFrame, index: Index) -> DataFrame:
        return replace(df, index, replacements)

    return _replace


def find_and_replace(
    df: DataFrame,
    exact_match: MatcherT,
    replacements: Iterable[Replacement],
) -> DataFrame:
    """Find exact matches and replace the values in those rows with replacements."""
    return replace(df, find(df, exact_match).index, replacements)


def find_and_replacer(
    exact_match: MatcherT,
    replacements: Iterable[Replacement],
) -> Callable[[DataFrame], DataFrame]:
    """Make a function to find exact matches and replace the values in those rows with replacements."""

    def _find_and_replace(df: DataFrame):
        return find_and_replace(df, exact_match, replacements)

    return _find_and_replace



from gzip import decompress as _decompress
from json import loads as _loads
from typing import Any as _Any
from typing import Dict as _Dict
from urllib.parse import urljoin as _urljoin

from requests import get as _get

from .importer import Importer as _Importer


class GunzipImporter(_Importer):
    def __init__(self: "GunzipImporter", domain: str) -> None:
        self.domain = domain

    def get(self: "GunzipImporter", url: str) -> _Dict[str, _Any]:
        response = _get(_urljoin(self.domain, url), verify=False)
        data = _decompress(response.content)
        ret: _Dict[str, _Any] = _loads(data)
        return ret



from .core import (
    DatasetInfo,
    ExactMatch,
    Fetcher,
    FileReader,
    FileWriter,
    FindCallBack,
    GunzipImporter,
    Importer,
    Replacement,
    Vault,
    find,
    find_and_replace,
    find_and_replacer,
    finder,
    fload,
    get_dataset_files,
    get_logger,
    list_datasets,
    reader_folder_path,
    replace,
    replacer,
    root_folder_path,
    set_logging_to_debug,
    set_logging_to_info,
    set_logging_to_warn,
    vault_token,
    writer_folder_path,
)

__all__ = [
    "DatasetInfo",
    "ExactMatch",
    "Fetcher",
    "FileReader",
    "FileWriter",
    "find_and_replace",
    "find_and_replacer",
    "find",
    "FindCallBack",
    "finder",
    "fload",
    "get_dataset_files",
    "get_logger",
    "GunzipImporter",
    "Importer",
    "list_datasets",
    "reader_folder_path",
    "replace",
    "Replacement",
    "replacer",
    "root_folder_path",
    "set_logging_to_debug",
    "set_logging_to_info",
    "set_logging_to_warn",
    "vault_token",
    "Vault",
    "writer_folder_path",
]



from abc import ABC as _ABC
from abc import abstractmethod as _abstractmethod

from ..fs.writer import WriterTypes as _WriterTypes


class Importer(_ABC):
    @_abstractmethod
    def get(self: "Importer", url: str) -> _WriterTypes:
        pass



from datetime import datetime as _datetime
from json import dump as _dump
from pathlib import Path as _Path
from typing import Any as _Any
from typing import Optional as _Optional
from typing import Type as _Type
from typing import TypeVar as _TypeVar
from typing import Union as _Union

from pandas import DataFrame as _DataFrame

from ..environment import writer_folder_path as _writer_folder_path
from ..logger import get_logger as _get_logger

T = _TypeVar("T", bound="FileWriter")
LOGGER = _get_logger()
WriterTypes = _Union[_DataFrame, _Any]


class FileWriter:
    def __init__(self: "FileWriter", folder: str, timestamp_col: _Optional[str] = None) -> None:
        self._timestamp = _datetime.utcnow()
        self._folder = _Path(folder)
        self._folder.mkdir(parents=True, exist_ok=True)
        self._timestamp_col = timestamp_col

    def write(self: "FileWriter", name: str, data: _Any) -> None:
        LOGGER.info("Writing file '%s'", name)
        if isinstance(data, _DataFrame):
            self.write_pandas(name, data)
        else:
            self.write_json(name, data)

    def write_pandas(self: "FileWriter", name: str, data: _DataFrame) -> None:
        file = self._get_file(name, "data.parquet")
        if self._timestamp_col:
            data[self._timestamp_col] = self._timestamp
        LOGGER.debug(f"Writing file '{file}'.")
        data.to_parquet(file)

    def write_json(self: "FileWriter", name: str, data: _Any) -> None:
        file = self._get_file(name, "output.json")
        if self._timestamp_col:
            if not isinstance(data, dict):
                data = {"data": data}
            data[self._timestamp_col] = self._timestamp.isoformat()
        LOGGER.debug(f"Writing file '{file}'.")
        with open(file, "w+") as output:
            _dump(data, output)

    @classmethod
    def from_environment(cls: _Type[T], **kwargs: _Any) -> T:
        return cls(_writer_folder_path(), **kwargs)

    def _get_file(self: "FileWriter", data_name: str, file_name: str) -> _Path:
        file = (
            self._folder.joinpath(data_name)
            .joinpath(self._timestamp.strftime("%Y-%m-%d/%H_%M_%S"))
            .joinpath(file_name)
        )
        file.parent.mkdir(parents=True, exist_ok=True)
        return file



from .datasets import DatasetInfo, get_dataset_files, list_datasets
from .edit import (
    ExactMatch,
    FindCallBack,
    Replacement,
    find,
    find_and_replace,
    find_and_replacer,
    finder,
    replace,
    replacer,
)
from .environment import reader_folder_path, root_folder_path, vault_token, writer_folder_path
from .fetch import Fetcher
from .fs import FileReader, FileWriter, fload
from .importer import GunzipImporter, Importer
from .logger import get_logger, set_logging_to_debug, set_logging_to_info, set_logging_to_warn
from .vault import Vault

__all__ = [
    "DatasetInfo",
    "ExactMatch",
    "Fetcher",
    "FileReader",
    "FileWriter",
    "find_and_replace",
    "find_and_replacer",
    "find",
    "FindCallBack",
    "finder",
    "fload",
    "get_dataset_files",
    "get_logger",
    "GunzipImporter",
    "Importer",
    "list_datasets",
    "reader_folder_path",
    "replace",
    "Replacement",
    "replacer",
    "root_folder_path",
    "set_logging_to_debug",
    "set_logging_to_info",
    "set_logging_to_warn",
    "vault_token",
    "Vault",
    "writer_folder_path",
]



from typing import Any as _Any
from typing import Dict as _Dict

from hvac import Client as _Client

from .environment import vault_token as _vault_token


class Vault:
    client = _Client(url="http://localhost:8200", token=_vault_token())

    def secret(self: "Vault", name: str) -> _Dict[str, _Any]:
        ret: _Dict[str, _Any] = self.client.read(f"kv/data/{name}")["data"]["data"]
        return ret



from collections import defaultdict
from pathlib import Path
from typing import Dict, List, Tuple

from brds.core.environment import reader_folder_path


def get_dataset_files(dataset_name: str) -> List[Tuple[str, List[Path]]]:
    root = Path(reader_folder_path())
    dataset_path = root.joinpath(dataset_name)
    grouped_files: Dict[str, List[Path]] = defaultdict(list)

    for file_path in dataset_path.glob("**/*.*"):
        if not file_path.is_file():
            continue
        timestamp_str = str(file_path.parent)
        grouped_files[timestamp_str].append(file_path.relative_to(root))

    # Sort by timestamp descending
    sorted_grouped_files = sorted(grouped_files.items(), key=lambda x: x[0], reverse=True)
    return sorted_grouped_files



from collections import defaultdict
from datetime import datetime
from os.path import isfile
from pathlib import Path
from typing import Any, Dict, List

from ..environment import reader_folder_path
from .dataset_info import DatasetInfo


def parse_timestamp(timestamp_str: str) -> datetime:
    try:
        return datetime.strptime(timestamp_str, "%Y-%m-%d/%H_%M_%S")
    except ValueError as ve:
        raise RuntimeError(f"Can't parse '{timestamp_str}'") from ve


def extract_dataset_info(dataset_path: Path, root: Path) -> Dict[str, Any]:
    relative_path = dataset_path.relative_to(root)
    parts = str(relative_path).split("/")
    date_str, time_str, _ = parts[-3:]
    dataset_name = parts[-4]
    module_name_parts = parts[:-4]
    module_name = "/".join(module_name_parts)
    full_name = f"{module_name}/{dataset_name}" if module_name else dataset_name
    timestamp = parse_timestamp(f"{date_str}/{time_str}")
    file_size = dataset_path.stat().st_size

    return {"full_name": full_name, "timestamp": timestamp, "file_size": file_size}


def list_datasets() -> List[DatasetInfo]:
    root = Path(reader_folder_path())
    dataset_list: List[DatasetInfo] = []

    dataset_versions: Dict[str, int] = defaultdict(int)
    dataset_data: Dict[str, Dict[str, Any]] = defaultdict(dict)

    for dataset_path in root.glob("**/*.*"):
        if not isfile(dataset_path):
            continue

        dataset_info = extract_dataset_info(dataset_path, root)
        full_name = dataset_info["full_name"]

        if full_name not in dataset_data or dataset_info["timestamp"] > dataset_data[full_name]["timestamp"]:
            dataset_data[full_name] = {"timestamp": dataset_info["timestamp"], "file_size": dataset_info["file_size"]}
        dataset_versions[full_name] += 1

    for full_name, data in dataset_data.items():
        if "/" in full_name:
            module_name, dataset_name = full_name.rsplit("/", 1)
        else:
            module_name = ""
            dataset_name = full_name
        dataset = DatasetInfo(
            module=module_name,
            name=dataset_name,
            timestamp=data["timestamp"],
            file_size=data["file_size"],
            old_versions=dataset_versions[full_name],
        )
        dataset_list.append(dataset)

    # Sort datasets by most recent modification time
    dataset_list.sort(key=lambda ds: ds.timestamp, reverse=True)
    return dataset_list



from pathlib import Path as _Path
from typing import Any, Dict

import pandas as pd
from fastapi import FastAPI, HTTPException, Path, Request, Response
from fastapi.responses import FileResponse, HTMLResponse
from fastapi.templating import Jinja2Templates

from brds import fload, get_dataset_files, list_datasets, reader_folder_path

app = FastAPI()

templates = Jinja2Templates(directory="./brds/templates")


@app.get("/dictionary/{filename:path}")
async def read_as_dict(filename: str = Path(..., regex=r"[\w\-/]+")) -> Dict[str, Any]:
    try:
        df = fload(filename)
        return df.to_dict(orient="records")
    except FileNotFoundError as exc:
        raise HTTPException(status_code=404, detail=f"Parquet file '{filename}' not found") from exc


@app.get("/raw/{filename:path}")
async def read_raw(filename: str = Path(..., regex=r"[\w\-/]+")) -> Dict[str, Any]:
    try:
        df = fload(filename)
        return df.to_dict(orient="records")
    except FileNotFoundError as exc:
        raise HTTPException(status_code=404, detail=f"Parquet file '{filename}' not found") from exc


@app.get("/html/{filename:path}", response_class=HTMLResponse)
async def read_html(filename: str = Path(..., regex=r"[\w\-/]+")) -> Response:
    try:
        df: pd.DataFrame = fload(filename)
        return df.to_html()
    except FileNotFoundError as exc:
        raise HTTPException(status_code=404, detail=f"Parquet file '{filename}' not found") from exc


@app.get("/datasets", response_class=HTMLResponse)
async def get_datasets(request: Request) -> Response:
    datasets = list_datasets()
    return templates.TemplateResponse("datasets.html", {"request": request, "modules": datasets})


@app.get("/download/{path:path}", response_class=FileResponse)
async def download_file(path: str):
    root = _Path(reader_folder_path())
    return FileResponse(root.joinpath(path), filename=path.split("/")[-1], media_type="application/octet-stream")


@app.get("/dataset/{dataset_name:path}", response_class=HTMLResponse)
async def dataset_files(request: Request, dataset_name: str):
    grouped_files = get_dataset_files(dataset_name)  # Implement this function to return grouped files
    return templates.TemplateResponse(
        "dataset_files.html", {"request": request, "dataset_name": dataset_name, "grouped_files": grouped_files}
    )



"""
Information about a dataset.
"""

from datetime import datetime

from brds.humanize import pretty_size


class DatasetInfo:
    """
    Information about a dataset.
    """

    def __init__(
        self: "DatasetInfo", module: str, name: str, timestamp: datetime, file_size: int, old_versions: int
    ) -> None:
        self.module = module
        self.name = name
        self.timestamp = timestamp
        self.file_size = file_size
        self.old_versions = old_versions

    def __repr__(self: "DatasetInfo") -> str:
        """
        String representation of the dataset info.

        >>> DatasetInfo("module", "name", datetime(2020, 1, 1), 0, 0)
        <DatasetInfo(module='module', name='name', timestamp=2020-01-01 00:00:00, file_size=0, old_versions=0)>
        """
        return (
            f"<DatasetInfo(module='{self.module}', name='{self.name}', "
            f"timestamp={self.timestamp}, file_size={self.file_size}, "
            f"old_versions={self.old_versions})>"
        )

    def __str__(self: "DatasetInfo") -> str:
        """
        String representation of the dataset info.

        >>> DatasetInfo("module", "name", datetime(2020, 1, 1), 0, 0)
        <DatasetInfo(module='module', name='name', timestamp=2020-01-01 00:00:00, file_size=0, old_versions=0)>
        """
        return self.__repr__()

    @property
    def pretty_file_size(self: "DatasetInfo") -> str:
        """
        Pretty print the file size.

        >>> DatasetInfo("module", "name", datetime(2020, 1, 1), 0, 0).pretty_file_size
        '0 B'
        >>> DatasetInfo("module", "name", datetime(2020, 1, 1), 1024, 0).pretty_file_size
        '1.00 KB'
        """
        return pretty_size(self.file_size)



from .dataset_files import get_dataset_files
from .dataset_info import DatasetInfo
from .list_datasets import list_datasets

__all__ = [
    "DatasetInfo",
    "get_dataset_files",
    "list_datasets",
]



from json import load as _load
from os import listdir as _listdir
from os.path import join as _join
from typing import Any as _Any
from typing import Optional as _Optional
from typing import TextIO as _TextIO
from typing import Type as _Type
from typing import TypeVar as _TypeVar

from pandas import read_parquet as _read_parquet

from ..environment import reader_folder_path as _reader_folder_path
from ..logger import get_logger as _get_logger

T = _TypeVar("T", bound="FileReader")

LOGGER = _get_logger()


class FileReader:
    def __init__(self: "FileReader", folder: str, version: _Optional[str] = None) -> None:
        self._root_folder = folder

        if version is None:
            self._folder = _last_folder(_last_folder(folder))
            self._version = self._folder[len(folder) :]
        else:
            self._folder = _join(folder, version)
            self._version = version

    def open(
        self: "FileReader",
        filename: _Optional[str] = None,
        *args: _Any,
        **kwargs: _Any,
    ) -> _TextIO:
        ret: _TextIO = open(self.get(filename), *args, **kwargs)
        return ret

    def get(self: "FileReader", filename: _Optional[str] = None) -> str:
        if filename is None:
            return _join(self._folder, _listdir(self._folder)[0])
        return _join(self._folder, filename)

    @classmethod
    def from_environment(cls: _Type[T], subfolder: str) -> T:
        return cls(_join(_reader_folder_path(), subfolder))

    def load(self: "FileReader", filename: _Optional[str] = None) -> _Any:
        new_file_name = self.get(filename)
        if filename:
            LOGGER.debug(
                "Latest file for folder '%s' (with filename='%s') resolved to '%s'.",
                self._root_folder,
                filename,
                new_file_name,
            )
        else:
            LOGGER.debug(
                "Latest file for folder '%s' resolved to '%s'.",
                self._root_folder,
                new_file_name,
            )
        if new_file_name.endswith(".json"):
            with open(new_file_name) as input_file:
                return _load(input_file)
        if new_file_name.endswith(".parquet"):
            return _read_parquet(new_file_name)
        raise NotImplementedError(f"Do not know how to load the file `{filename}`: `{new_file_name}`")


def _last_folder(folder: str) -> str:
    return _join(folder, sorted(_listdir(folder), reverse=True)[0])


def fload(folder: str, filename: _Optional[str] = None) -> _Any:
    if not filename:
        LOGGER.debug("Loading the file '%s'", folder)
    else:
        LOGGER.debug("Loading the file '%s/%s'", folder, filename)
    return FileReader.from_environment(folder).load(filename)
